var elxwchatgui = null;

if (typeof String.prototype.format == 'undefined') {
    String.prototype.format = function()
    {
       var content = this;
       for (var i=0; i < arguments.length; i++)
       {
            var replacement = '{' + i + '}';
            content = content.replace(replacement, arguments[i]);
       }
       return content;
    };
}

$( document ).ready(function() {
    elxwchatgui = new ElastixSIPPanelGUI('div#elastix-panel-elastixwebphone > div.panel-body');
    elxwchatgui.setupSIPClient();
});


/* La siguiente clase representa el GUI del panel de llamada y chat SIP. */
function ElastixSIPPanelGUI(selector)
{
    // Conexión actual SIP WebRTC
    this._ua = null;

    // Objeto que maneja presencia SIP
    this._sp = null;

    // Sesión de llamada activa
    this._session = null;

    // Subscripción a Message Waiting Indication
    this._mwi = null;

    // Objeto que maneja BLF
    this._blf = null;

    // Panel dentro del cual se contiene todo lo demás
    this.panel = $(selector);

    // Reproductor de DTMFs
    this._dtmf = new PlayDTMF();

    // Cambio de pestañas entre los objetivos
    this._panelAS().find('div#bar-buttons > a.barnav').click(function(e) {
        e.preventDefault();
        var a = $(e.currentTarget);
        var href = a.attr('href');

        a.parent('div#bar-buttons').children('a.barnav').removeClass('selected');
        a.addClass('selected');
        this._panelAS().find('div.phonetab').hide();
        this._panelAS().find('div.phonetab'+href).show();
    }.bind(this));

    // Pulsar botón en keypad
    this._panelAS().find('div#keypad > div.phonekey').click(function(e) {
        var div = $(e.currentTarget);
        var key = div.data('key');
        var sess = this.getCurrSession();
        if (sess != null) sess.dtmf(key);
        var dialtxt = this._panelAS().find('div#phonebox > input#elxwebphone-dialed-extension');
        dialtxt.val(dialtxt.val() + key);
        this._dtmf.play(key);
    }.bind(this));

    // Filtrado de mensajes y activación de contactos offline
    this._panelAS().find('div#contacts > div#elx_im_contact_search > input#im_search_filter')
        .keyup(this._updateContactVisibility.bind(this));
    this._panelAS().find('div#contacts #elx-chk-show-offline-contacts')
        .click(this._updateContactVisibility.bind(this));

    // Clic en contacto copia la extensión para iniciar una llamada o chat
    this._panelAS().find('div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts')
        .on('click', 'li', function(e) {
        var data = $(e.currentTarget).data();
        var dialtxt = this._panelAS().find('div#phonebox > input#elxwebphone-dialed-extension');
        dialtxt.val(data.extension);

        // Cambiar a chat activo si existe
        this._selectChat(data.uri, false);

        // Foco de selección en el input de teléfono
        dialtxt.focus();
    }.bind(this));

    // ENTER en el input de teléfono inicia llamada
    this._panelAS().find('div#phonebox > input#elxwebphone-dialed-extension').keydown(function (event) {
        var dial = $(event.currentTarget).val();

        if (event.which == 13) {
            event.preventDefault();
            if (dial != '') {
                this._panelAS().find('div#bar-buttons > a#elxwebphone-do-call').click();

                // Ignore Enter when empty input.
            } else return false;
        }
    }.bind(this));

    // Marcar llamada
    this._panelAS().find('div#bar-buttons > a#elxwebphone-do-call')
        .click(function() {
        var ua = this._getCurrSIPUA();
        var sess = this.getCurrSession();
        if (sess != null) return;   // Esto no debería pasar

        var dialtxt = this._panelAS().find('div#phonebox > input#elxwebphone-dialed-extension');
        var dialstring = dialtxt.val();
        dialtxt.val('');

        // Construir cadena de marcado completa con dominio
        dialstring = dialstring.trim();
        if (dialstring == '') return;
        if (dialstring.indexOf('@') == -1) {
            dialstring += '@' + ua.configuration.hostportParams;
        }

        sess = ua.invite(dialstring, this._getVideoCallMedia());
        this.setCurrSession(sess);

        this.displayPhoneOnCall();
        this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_DIALING + ': ' + dialstring);
    }.bind(this));

    // Iniciar chat
    this._panelAS().find('div#bar-buttons > a#elxwebphone-do-chat')
        .click(function() {
        var ua = this._getCurrSIPUA();

        var dialtxt = this._panelAS().find('div#phonebox > input#elxwebphone-dialed-extension');
        var dialstring = dialtxt.val();

        // Construir cadena de marcado completa con dominio
        dialstring = dialstring.trim();
        if (dialstring == '') return;
        if (dialstring.indexOf('@') == -1) {
            dialstring += '@' + ua.configuration.hostportParams;
        }

        this._selectChat(dialstring, true);
        this._panelAS().find('div#chats > textarea').focus();
    }.bind(this));

    // Seleccionar chat activo a través del select
    this._panelAS().find('div#chats > select#elx_chat_active')
        .change(function() {
        var sel = this._panelAS().find('div#chats > select#elx_chat_active');
        var uri = sel.val();
        var mainchat_ctl = 'div#chats > textarea, div#chats > span#elx_close_chat, div#chats > div#elx_chat_group';

        this._panelAS().find('div#chats > div#elx_chat_group > div').hide();
        this._panelAS().find('div#chats > div#elx_chat_group > div[data-uri="' + uri + '"]').show();
        if (uri == '') {
            this._panelAS().find(mainchat_ctl).hide();
        } else {
            this._panelAS().find(mainchat_ctl).show();
        }
    }.bind(this));

    // Eliminar el chat seleccionado en el select
    this._panelAS().find('div#chats > span#elx_close_chat').click(function() {
        var sel = this._panelAS().find('div#chats > select#elx_chat_active');

        this._closeChat(sel.val());
    }.bind(this));

    // Enviar mensaje de chat
    this._panelAS().find('div#chats > textarea').on('keydown', function(event) {
        var msg = $(event.currentTarget).val();

        // Ignore TAB and ESC.
        if (event.which == 9 || event.which == 27) {
            return false;
            // Enter pressed? so send chat.
        } else if (event.which == 13) {
            event.preventDefault();
            if (msg != '') {
                $(event.currentTarget).val('');
                var sel = this._panelAS().find('div#chats > select#elx_chat_active');
                this._sendMessage(msg, sel.val());

                // Ignore Enter when empty input.
            } else return false;
        }
    }.bind(this));

    // Contestar llamada
    this._panelAS().find('div#bar-buttons > a#elxwebphone-do-answer')
        .click(function() {
        var sess = this.getCurrSession();
        if (sess != null) sess.accept(this._getVideoCallMedia());
    }.bind(this));

    // Rechazar llamada
    this._panelAS().find('div#bar-buttons > a#elxwebphone-do-reject')
        .click(function() {
        var sess = this.getCurrSession();
        if (sess != null) sess.reject();
    }.bind(this));

    // Colgar/cancelar llamada
    this._panelAS().find('div#bar-buttons > a#elxwebphone-do-hangup')
        .click(function() {
        var sess = this.getCurrSession();
        if (sess != null) {
            if (sess.startTime != null)
                sess.bye();
            else
                sess.cancel();
        }
    }.bind(this));
}
/**
 * Procedimiento que inicia la configuración de las configuraciones SIP. Las
 * dependencias que hay que resolver hasta ahora son las siguientes:
 *
 * Carga credenciales SIP,
 *  requiere: (nada)
 * Inicio cliente SIP
 *  requiere: Carga credenciales SIP
 * Carga lista de contactos
 *  requiere: (nada)
 * Creación de lista de contactos con data
 *  requiere: Carga lista de contactos
 * Publicación de presencia propia
 *  requiere: Inicio cliente SIP
 * Subscripción a lista de contactos
 *  requiere: Inicio cliente SIP, Creación de lista de contactos con data
 *
 * @return void
 */
ElastixSIPPanelGUI.prototype.setupSIPClient = function ()
{
    this.displaySessionLoading();

    var deferred_registerUA = new $.Deferred();
    var deferred_rosterLoad = new $.Deferred();

    // Iniciar la carga de las credenciales SIP
    $.get('index.php', {
        menu: '_elastixutils',
        action: 'extension_current_user'
    }, function (response) {
        if (response.error != '') {
            // Rechazar la promesa con el mensaje de error
            deferred_registerUA.reject(response.error);
        } else if (response.message.tech != 'sip') {
            // Tecnología no es SIP
            deferred_registerUA.reject(arrLang_panel.elastixwebphone.MSG_INCOMPATIBLE_VOIP_TECH);
        } else {
            var arrData = response.message;

            // Con las credenciales, se puede iniciar el SIP.UA
            var ip_pbx = window.location.host;
            this._ua = new SIP.UA({
                //log:                {level: "debug"},
                //traceSip:           true,
                uri:                arrData.authorizationUser + '@' + ip_pbx,
                wsServers:          'wss://' + ip_pbx + '/asterisk/ws',
                //displayName:        response.message.display_name,
                authorizationUser:  arrData.authorizationUser,
                password:           arrData.password,
                hackIpInContact:    true,
                hackWssInTransport: true,
                autostart:          true
                //register:           response.message.register,
                //registerExpires:    response.message.register_expires,
                //usePreloadedRoute:  response.message.use_preloaded_route,
                //noAnswerTimeout:    response.message.no_answer_timeout,
                //stunServers:        ["stun:null"],
            });
            this._ua.on('message', function (e) {
                var remoteUri = e.remoteIdentity.uri.toString();
                var elx_txt_chat = e.body;
                var remoteUser = remoteUri.split('sip:');
                var uri2 = remoteUser[1];

                this.receiveMessage(uri2, e.contentType, elx_txt_chat);
            }.bind(this)).on('disconnected', function() {
                this.displayStatusError(arrLang_panel.elastixwebphone.MSG_DISCONNECTED);
            }.bind(this)).on('registered', function () {
                // El registro exitoso es uno de los requisitos para iniciar suscripción
                this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_READY_DIAL);
                deferred_registerUA.resolve();

                this.subscribeMWI();
                this._blf = new SIPBLF(this._ua, this);
            }.bind(this)).on('registrationFailed', function(cause) {
                deferred_registerUA.reject(arrLang_panel.elastixwebphone.MSG_REGISTRATION_FAILED + ': ' + cause);
            }.bind(this)).on('invite', function (incoming_session) {
                // Rechazar llamada si hay otra activa.
                if (this.getCurrSession() != null) {
                    incoming_session.reject();
                    return;
                }
                this.setCurrSession(incoming_session);
                this.displayPhoneIncoming(
                    incoming_session.remoteIdentity.uri.user,
                    incoming_session.remoteIdentity.displayName);
            }.bind(this));
        }
    }.bind(this)).fail(function() {
        deferred_registerUA.reject(arrLang_panel.elastixwebphone.MSG_SIP_PARAM_FAILED);
    }.bind(this));

    // Iniciar la carga de la lista de contactos
    $.getJSON( "rest.php/address_book/ContactList/internal", function(data) {
        this.clearContactRoster();
        $.each( data.extension, function(key, val) {

            var uri = val.user + '@' + window.location.host;
            this.addContact2Roster(val.id, val.name, uri, uri, val.phone, val);
        }.bind(this));
        deferred_rosterLoad.resolve();
    }.bind(this)).fail(function() {
        deferred_rosterLoad.reject(arrLang_panel.elastixwebphone.MSG_SIP_ROSTER_FAILED);
    }.bind(this));

    // Iniciar la subscripción cuando ambas promesas puedan resolverse
    $.when(deferred_registerUA, deferred_rosterLoad)
    .done(function() {
        this.displaySessionActive();

        this._sp = new SIPPresence(this._ua, this);
        this._sp.publishPresence();
    }.bind(this)).fail(function(msg) {
        this.displaySessionError(msg);
    }.bind(this));
}
ElastixSIPPanelGUI.prototype._getCurrSIPUA = function() {
    return this._ua;
}
ElastixSIPPanelGUI.prototype.getCurrSession = function() {
    return this._session;
}
ElastixSIPPanelGUI.prototype.displaySessionLoading = function() {
    // Marcar color de presencia como desconectado hasta actualizar
    this.updateMyStatus('gray');
    this.panel.find('div.elxwebphone-loadstage').hide();
    this.panel.find('div#elxwebphone-starting-session').show();
}
ElastixSIPPanelGUI.prototype.displaySessionActive = function() {
    this.panel.find('div.elxwebphone-loadstage').hide();
    this.panel.find('div#elxwebphone-active-session').show();

    this._updateContactVisibility();
}
ElastixSIPPanelGUI.prototype.displaySessionError = function(msg) {
    this.panel.find('div.elxwebphone-loadstage').hide();
    this.panel.find('div#elxwebphone-error-session > span').text(msg);
    this.panel.find('div#elxwebphone-error-session').show();
}
ElastixSIPPanelGUI.prototype._panelAS = function() {
    return this.panel.children('div#elxwebphone-active-session');
}
ElastixSIPPanelGUI.prototype.displayStatusMessage = function (statustxt) {
    this._panelAS().find('div#phonebox > div#dial-statustext')
        .removeClass('error')
        .text(statustxt);
}
ElastixSIPPanelGUI.prototype.displayStatusError = function (statustxt) {
    this._panelAS().find('div#phonebox > div#dial-statustext')
        .addClass('error')
        .text(statustxt);
}
ElastixSIPPanelGUI.prototype.clearContactRoster = function() {
    this._panelAS().find('div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts')
        .empty();
}
ElastixSIPPanelGUI.prototype.addContact2Roster = function(idUser, display_name, uri, alias,
    extension, extradata) {
    var listcontacts = this._panelAS().find('div#contacts > div#elx_im_list_contacts');
    var liContact = listcontacts.find('ul#elx_template_contact_status > li').clone()
        .attr('data-status', 'offline')
        .attr('data-uri', uri)
        .attr('data-alias', alias)
        .attr('data-name', display_name)
        .attr('data-idUser', idUser)
        .attr('data-extension', extension);
    liContact.find('div.elx_contact_div').text(display_name);
    this._updateContactTag(liContact, 'grey', '(unknown)');
    listcontacts.children('ul#elx_ul_list_contacts').append(liContact);
}
ElastixSIPPanelGUI.prototype.updateContactStatus = function(contact, newColor, newNote) {
    var liContact = this._panelAS().find(
        "div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li[data-uri='" + contact + "']");
    this._updateContactTag(liContact, newColor, newNote);
    this._updateContactVisibility();
}
ElastixSIPPanelGUI.prototype._updateContactTag = function (liContact, newColor, newNote) {
    liContact.data('status', (newColor != 'grey') ? 'online' : 'offline');
    liContact.find('i.fa-user').css('color', newColor);
    liContact.attr('title', newNote);
}
ElastixSIPPanelGUI.prototype.updateMyStatus = function(newColor) {
    // TODO: mostrar color de estado del propio usuario
    //$(".elx-content-photo").css('border-color', newColor);
}
ElastixSIPPanelGUI.prototype.updateContactBLF = function(contact, status) {
    var liContact = this._panelAS().find(
        "div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li[data-uri='" + contact + "']");

    liContact.children('i.fa-phone')
        .removeClass('ringing')
        .removeClass('oncall');
    if (status == 'ringing' || status == 'oncall')
        liContact.children('i.fa-phone').addClass(status);
}
/**
 * Esta función actualiza el estado de visibilidad de todos los contactos. Un
 * contacto debe mostrarse si su nombre coincide con el patrón de búsqueda (una
 * cadena vacía coincide con todo), y si deben mostrarse los contactos offline,
 * o si el contacto en sí está online.
 */
ElastixSIPPanelGUI.prototype._updateContactVisibility = function() {
    var pattern = this.panel.find("input#im_search_filter").val();
    var mostrarTodos = this.panel.find('#elx-chk-show-offline-contacts').is(':checked');

    // Ocultar todos los contactos, y volver a mostrar sólo coincidencias
    this._panelAS().find("div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li")
    .hide()
    .each(function() {
        var n = $(this).data('name');
        n = n.toString();   // Obligar a que sea cadena, no número
        if (n.match(pattern) &&
            (mostrarTodos || $(this).data('status') != 'offline')) {
            $(this).show();
        }
    });
}
ElastixSIPPanelGUI.prototype.getContactURIs = function() {
    return this._panelAS().find("div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li")
        .map(function (i, e) { return $(e).data('uri'); })
        .get();
}
ElastixSIPPanelGUI.prototype.displayPhoneIdle = function() {
    // Mostrar botones que corresponden al estado ocioso
    this._panelAS().find('div#bar-buttons > a.elxwebphone-action')
        .hide();
    this._panelAS().find('div#bar-buttons > a.elxwebphone-onidle')
        .show();
    this._panelAS().find('video#elx_video_remote, div#elxwebphone-active-session > video#elx_video_local')
        .hide();
}
ElastixSIPPanelGUI.prototype.displayPhoneIncoming = function(dialstring, displayname) {
    // Mostrar botones que corresponden a llamada entrante que timbra
    this._panelAS().find('div#bar-buttons > a.elxwebphone-action')
        .hide();
    this._panelAS().find('div#bar-buttons > a.elxwebphone-onring')
        .show();

    this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_INCOMING_CALL.format(displayname, dialstring));
}
ElastixSIPPanelGUI.prototype.displayPhoneOnCall = function() {
    // Mostrar botones que corresponden a llamada activa
    this._panelAS().find('div#bar-buttons > a.elxwebphone-action')
        .hide();
    this._panelAS().find('div#bar-buttons > a.elxwebphone-oncall')
        .show();
    // Desactivar video hasta hacer pruebas sobre ubicación de tag <video> en página

    var showvideo = false;
    var sess = this.getCurrSession();
    var ms = sess.getRemoteStreams();
    if (ms.length > 0) {
        var vt = ms[0].getVideoTracks();
        if (vt.length > 0) showvideo = true;
    }

    if (showvideo) {
        this._panelAS().find('video#elx_video_remote, div#elxwebphone-active-session > video#elx_video_local')
            .show();
    }
}
ElastixSIPPanelGUI.prototype._getVideoCallMedia = function() {
    return {
        media: {
            constraints: {
                audio: true,
                video: true
            },
            render: {
                remote: this._panelAS().find('#elx_video_remote').get(0),
                local: this._panelAS().find('#elx_video_local').get(0)
            }
        }
    };
}
ElastixSIPPanelGUI.prototype.setCurrSession = function(sess) {
    this._session = sess;
    if (sess == null) return;

    sess.on('progress', function(response) {
        this.displayStatusMessage(response.status_code + ' ' + response.reason_phrase);//alert-info
    }.bind(this)).on('cancel', function() {
        this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_CALL_CANCELLED);//alert-success
        this.setCurrSession(null);
        this.displayPhoneIdle();
    }.bind(this)).on('accepted', function(data) {
        this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_CALL_ACCEPTED);//alert-success
        this.displayPhoneOnCall();
    }.bind(this)).on('failed', function(response, cause) {
        var msg = cause +
            ((response != null && (typeof response != 'string'))
                ? (': ' + response.status_code + ' ' + response.reason_phrase)
                : '');
        this.displayStatusError(msg);//alert-danger
        this.setCurrSession(null);
        this.displayPhoneIdle();
    }.bind(this)).on('bye', function(request) {
        this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_CALL_TERMINATED);//alert-info
        this.setCurrSession(null);
        this.displayPhoneIdle();
    }.bind(this)).on('terminated', function(message, cause) {
        this.displayStatusMessage(arrLang_panel.elastixwebphone.MSG_CALL_TERMINATED);//alert-info
        this.setCurrSession(null);
        this.displayPhoneIdle();
    }.bind(this));
}
ElastixSIPPanelGUI.prototype._selectChat = function(uri, create) {
    var sel = this._panelAS().find('div#chats > select#elx_chat_active');

    // Buscar si existe el chat activo para el uri indicado
    var op = sel.children('option[value="' + uri + '"]');
    if (op.length <= 0) {
        if (!create) return;

        this._createChat(uri);
    }

    // Simular que se elige el tab de chat
    this._panelAS().find('div#bar-buttons > a#barnav-chats').click();

    // Seleccionar el select
    sel.val(uri);
    sel.change();
}
ElastixSIPPanelGUI.prototype._createChat = function(uri) {
    var sel = this._panelAS().find('div#chats > select#elx_chat_active');
    var chatgroup = this._panelAS().find('div#chats > div#elx_chat_group');

    // Buscar nombre descriptivo en la lista de contactos
    var chatLabel = uri;
    var liContact = this._panelAS().find("div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li[data-uri=\"" + uri + "\"]");
    if (liContact.length > 0) chatLabel = liContact.data('name');

    // Crear el item de select y el div
    $('<option />')
        .val(uri)
        .text(chatLabel)
        .appendTo(sel);
    $('<div style="display: none;"/>')
        .attr('data-uri', uri)
        .appendTo(chatgroup);
    sel.children('option[value=""]').hide();
    sel.prop('disabled', false);

    // Si el valor previo era el de no-chat, mostrar el recién agregado
    if (sel.val() == '') {
        sel.val(uri);
        sel.change();
    }
}
ElastixSIPPanelGUI.prototype._sendMessage = function (msg_txt, uri) {
    var ua = this._getCurrSIPUA();
    this._addMessage(uri, 'out', msg_txt);

    ua.message(uri, msg_txt).on('failed', function (response, cause) {
        var error_msg = (response)
            ? response.status_code.toString() + " " + response.reason_phrase
            : cause;

        var target = this._panelAS().find('div#chats > div#elx_chat_group > div[data-uri="' + uri + '"]');
        target.find("div").last().css("color","red");
        this._addMessage(uri, 'in', $('<span style="color: red;"></span>').text(error_msg));
    });
}
ElastixSIPPanelGUI.prototype.receiveMessage = function(urioralias, content_type, msg_txt) {
    /* En Elastix 2 y 4, el valor de urioralias puede tener un dominio con una
     * IP, ya que se puede recibir el mensaje desde un endpoint conectado en
     * una interfaz distinta. Se separa el valor de extensión y se busca en el
     * roster SIP para normalizar, antes de buscar entre los chats activos.  */
    var m = /^(sip:)?(\S+)@(\S+)$/.exec(urioralias);
    var user = m[2];
    var domain = m[3];
    var liContact = this._panelAS().find('div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li[data-extension="'+user+'"]');
    if (liContact.length > 0) {
        urioralias = liContact.data('uri');
    }

    var sel = this._panelAS().find('div#chats > select#elx_chat_active');
    // Buscar si existe el chat activo para el uri indicado
    var op = sel.children('option[value="' + urioralias + '"]');
    if (op.length <= 0) {
        this._createChat(urioralias);
    }

    if (content_type == 'application/im-iscomposing+xml') {
        var xml = $.parseXML(msg_txt);
        var state = $(xml).find('isComposing > state').text();
        if (state != 'active') state = 'idle';

        this._setComposingState(urioralias, state);
    } else {
        this._addMessage(urioralias, 'in', msg_txt);
    }
}
ElastixSIPPanelGUI.prototype._setComposingState = function(uri, state) {
    var target = this._panelAS().find('div#chats > div#elx_chat_group > div[data-uri="' + uri + '"]');
    target.find('div.elx_chat_composing').remove();
    if (state == 'active') {
        var messagediv = $('<div class="elx_chat_composing"></div>');
        messagediv.text(arrLang_panel.elastixwebphone.MSG_CHAT_COMPOSING);
        target.append(messagediv).scrollTop(1e4);
    }
}
ElastixSIPPanelGUI.prototype._addMessage = function(uri, direction, message) {
    var send_name;
    if (direction == 'out') {
        send_name = arrLang_panel.elastixwebphone.MSG_CHAT_MYSELF;
    } else {
        send_name = uri;
        var liContact = this._panelAS().find("div#contacts > div#elx_im_list_contacts > ul#elx_ul_list_contacts > li[data-uri=\"" + uri + "\"]");
        if (liContact.length > 0) send_name = liContact.data('name');
    }

    // Para ahorrar espacio, se toma el primer elemento sin espacios
    var tokens = send_name.trim().split(' ');
    send_name = tokens[0];
    if (send_name == 'undefined' || send_name == '') send_name = 'receive';
    //if (direction != 'out') elxTitleAlert('New Message ' + send_name);

    var messagediv = $('<div></div>');
    if (typeof message == 'string')
        messagediv.text(message);
    else messagediv.append(message);

    var chatdiv = this._panelAS().find('div#chats > div#elx_chat_group > div[data-uri="' + uri + '"]');

    chatdiv
        .append(messagediv.prepend($('<b></b>').text(send_name + ': ')))
        .scrollTop(1e4);
    if (direction != 'out') {
        chatdiv.find('div.elx_chat_composing').remove();
    }

}
ElastixSIPPanelGUI.prototype._closeChat = function (uri) {
    var sel = this._panelAS().find('div#chats > select#elx_chat_active');

    // Buscar si existe el chat activo para el uri indicado
    if (uri == '') return;
    var op = sel.children('option[value="' + uri + '"]');
    if (op.length <= 0) return;

    // Quitar div y option del select
    this._panelAS().find('div#chats > div#elx_chat_group > div[data-uri="' + uri + '"]').remove();
    op.remove();

    // Si sólo queda el elemento de no-chats, mostrarlo y ocultar controles
    if (sel.children().length <= 1) {
        sel.children('option[value=""]').show();
    } else {
        // Elegir primer elemento no-vacío
        if (sel.val() == '') {
            var nval = sel.children('option[value!=""]').first().val();
            sel.val(nval);
        }
    }
    sel.change();
}
ElastixSIPPanelGUI.prototype.subscribeMWI = function() {
    if (this._mwi != null) return;

    var ua = this._getCurrSIPUA();
    var s_options = {
        expires:        120,
        extraHeaders:   [
            'Accept: application/simple-message-summary'
        ]
    };
    this._mwi = ua.subscribe(this.getLocalContact(), 'message-summary', s_options);
    this._mwi.on('notify', this._notifyMWI.bind(this));
}
ElastixSIPPanelGUI.prototype.unsubscribeMWI = function() {
    if (this._mwi == null) return;

    this._mwi.unsubscribe();
    this._mwi.close();
    this._mwi = null;
}
ElastixSIPPanelGUI.prototype._notifyMWI = function(notification) {
    var headers = {};
    var rx = /^(\d+)\/(\d+)(\s*\((\d+)\/(\d+)\))?$/;
    var mwi = this._panelAS().find('div#phonebox > div#mwi')

    mwi.children('span').hide();
    mwi.find('span > span.number').hide().text('0');

    notification.request.body
        .trim().split("\r\n").forEach(function(val, idx) {
        var f = val.split(": ", 2);
        var m = rx.exec(f[1]);
        if (m != null) {
            headers[f[0]] = [
                parseInt(m[1]),
                parseInt(m[2]),
                (m[4] == undefined) ? 0 : parseInt(m[4]),
                (m[5] == undefined) ? 0 : parseInt(m[5])
            ];
            if (headers[f[0]][0] > 0) {
                mwi.find('span#' + f[0] + ' > span.number').text(headers[f[0]][0]);
                if (headers[f[0]][0] > 1)
                    mwi.find('span#' + f[0] + ' > span.number').show();
                mwi.children('span#' + f[0]).show();
            }
        } else {
            headers[f[0]] = f[1];
        }
    });

    if (headers['Messages-Waiting'] == 'yes')
        mwi.show();
    else
        mwi.hide();
}
ElastixSIPPanelGUI.prototype.getLocalContact = function() {
    var ua = this._getCurrSIPUA();
    return ua.configuration.authorizationUser + '@' + ua.configuration.hostportParams;
}

/**
 * La clase Presentity es una clase que modela el documento XML que contiene la
 * información de presencia RPID. La presencia rica puede estar activa (propiedad
 * open en TRUE) o inactiva. Se puede agregar una nota para describir con más
 * detalle el tipo de actividad en que se encuentra el usuario. Además se soporta
 * una lista de actividades (propiedad activities) cuyo contenido determina si
 * el usuario está disponible, ocupado, ausente, u otra cosa.
 *
 * La clase soporta generar el XML a partir de las propiedades, y además parsear
 * un documento XML y separar sus propiedades. La implementación se ha probado
 * con el documento XML publicado por Jitsi.
 */
function Presentity()
{
    // urn:ietf:params:xml:ns:pidf      http://tools.ietf.org/html/rfc3863
    // urn:ietf:params:xml:ns:pidf:rpid http://tools.ietf.org/html/rfc4480
    this.user = "user";
    this.domain = "example.com";
    this.activities = [];
    this.status_icon = null;
    this.open = true;
    this.note = "Online";
}
/** Generación del documento XML a partir de las propiedades */
Presentity.prototype.toXML = function() {
    var xml_presence = $.parseXML(
        '<?xml version="1.0" encoding="UTF-8" standalone="no"?>' +
        '<presence xmlns="urn:ietf:params:xml:ns:pidf" xmlns:dm="urn:ietf:params:xml:ns:pidf:data-model" xmlns:rpid="urn:ietf:params:xml:ns:pidf:rpid"/>');
    var sipuri = 'sip:' + this.user + '@' + this.domain;

    /* A partir de aquí se usan funciones nativas de XMLDocument porque la
     * abstracción de jQuery no permite agregar elementos con namespace
     */
    var xmlpr = xml_presence.getElementsByTagName('presence')[0];
    xmlpr.setAttribute('entity', sipuri);

    var xmlpers = xml_presence.createElementNS("urn:ietf:params:xml:ns:pidf:data-model", "person");
    xmlpers.setAttribute('id', 'p1401');
    var xmlactivities = xml_presence.createElementNS("urn:ietf:params:xml:ns:pidf:rpid", "activities");

    // http://tools.ietf.org/html/rfc4480
    var knownactivities = ['appointment', 'away', 'breakfast', 'busy', 'dinner',
        'holiday', 'in-transit', 'looking-for-work', 'lunch', 'meal', 'meeting',
        'on-the-phone', 'performance', 'permanent-absence', 'playing', 'presentation',
        'shopping', 'sleeping', 'spectator', 'steering', 'travel', 'tv', 'unknown',
        'vacation', 'working', 'worship'];
    for (var i = 0; i < this.activities.length; i++) {
        var xmlactv
        if (-1 != knownactivities.indexOf(this.activities[i])) {
            xmlactv = xml_presence.createElementNS("urn:ietf:params:xml:ns:pidf:rpid", this.activities[i]);
        } else {
            xmlactv = xml_presence.createElementNS("urn:ietf:params:xml:ns:pidf:rpid", 'other');
            xmlactv.appendChild(xml_presence.createTextNode(this.activities[i]));
        }
        xmlactivities.appendChild(xmlactv);
    }
    xmlpers.appendChild(xmlactivities);
    if (this.status_icon != null) {
        var xmlicon = xml_presence.createElementNS("urn:ietf:params:xml:ns:pidf:rpid", "status-icon");
        xmlicon.appendChild(xml_presence.createTextNode(this.status_icon));
        xmlpers.appendChild(xmlicon);
    }
    xmlpr.appendChild(xmlpers);


    var xmltuple = xml_presence.createElement('tuple');
    xmltuple.setAttribute('id', 't1072');

    var xmlstatus = xml_presence.createElement('status');
    var xmlbasic = xml_presence.createElement('basic');
    xmlbasic.appendChild(xml_presence.createTextNode(this.open ? 'open' : 'closed'));
    xmlstatus.appendChild(xmlbasic);
    xmltuple.appendChild(xmlstatus);
    var xmlcontact = xml_presence.createElement('contact');
    xmlcontact.appendChild(xml_presence.createTextNode(sipuri));
    xmltuple.appendChild(xmlcontact);
    var xmlnote = xml_presence.createElement('note');
    xmlnote.appendChild(xml_presence.createTextNode(this.note));
    xmltuple.appendChild(xmlnote);
    xmlpr.appendChild(xmltuple);

    return xml_presence;
}

Presentity.prototype.fromXML_sipuri = function(sipuri) {
    var m = /^(sip:)?(\S+)@(\S+)$/.exec(sipuri);
    this.user = m[2];
    this.domain = m[3];
}

/** Parseo de un XML y extracción de las propiedades */
Presentity.prototype.fromXML = function(xml_presence) {
    var xmlpr = xml_presence.getElementsByTagName('presence')[0];

    // Blink manda el atributo entity codificado
    var sipuri = decodeURIComponent(xmlpr.getAttribute('entity'));
    this.fromXML_sipuri(sipuri);

    this.activities = [];
    var xmlactivities = xml_presence.getElementsByTagNameNS(
        "urn:ietf:params:xml:ns:pidf:rpid",
        //urn:ietf:params:xml:ns:pidf:rpid:rpid-person
        "activities");
    if (xmlactivities.length > 0) for (var i = 0; i < xmlactivities[0].childNodes.length; i++) {
        var xmlactv = xmlactivities[0].childNodes[i];
        if (xmlactv.nodeType == xml_presence.ELEMENT_NODE) {
            var m = /^(\S+:)?(\S+)/.exec(xmlactv.nodeName);
            var nodeName = m[2];

            if (nodeName == 'other') {
                // Asume que el texto es el único contenido
                if (xmlactv.childNodes.length > 0)
                    this.activities.push(xmlactv.childNodes[0].nodeValue);
            } else {
                this.activities.push(nodeName);
            }
        }
    }

    this.status_icon = null;
    var xmlicons = xml_presence.getElementsByTagNameNS(
        "urn:ietf:params:xml:ns:pidf:rpid",
        "status-icon");
    if (xmlicons.length > 0) {
        // Asume que el texto es el único contenido
        if (xmlicons[0].childNodes.length > 0)
            this.status_icon = xmlicons[0].childNodes[0].nodeValue;
    }

    // Asterisk pone el tag note aquí
    this.note = "Offline";
    var xmlnote = xml_presence.getElementsByTagName('note');
    if (xmlnote.length > 0) {
        this.note = xmlnote[0].childNodes[0].nodeValue;
    }

    var xmltuple = xml_presence.getElementsByTagName('tuple');
    if (xmltuple.length > 0) {

        // Asterisk manda el nombre del contacto aquí
        var xmlcontact = xmltuple[0].getElementsByTagName('contact');
        if (xmlcontact.length > 0) {
            sipuri = xmlcontact[0].childNodes[0].nodeValue;
            this.fromXML_sipuri(sipuri);
        }

        this.open = false;
        var xmlstatus = xmltuple[0].getElementsByTagName('status');
        if (xmlstatus.length > 0) {
            var xmlbasic = xmlstatus[0].getElementsByTagName('basic');
            if (xmlbasic.length > 0) {
                this.open = ('open' == xmlbasic[0].childNodes[0].nodeValue);
            }
        }

        var xmlnote = xmltuple[0].getElementsByTagName('note');
        if (xmlnote.length > 0) {
            this.note = xmlnote[0].childNodes[0].nodeValue;
        } else {
            // Blink no asigna la nota. Se sintetiza una en base a las actividades
            if (this.activities.indexOf('available') != -1) {
                this.note = 'Online';
            } else if (this.activities.indexOf('away') != -1) {
                this.note = 'Away';
            } else if (this.activities.indexOf('busy') != -1) {
                this.note = 'Busy (DND)';
            }
        }
    }
}

Presentity.prototype.toString = function() {
    var xml = (new XMLSerializer()).serializeToString(this.toXML());
    xml = xml.replace(/ xmlns=""/g, '');
    return xml;
}

Presentity.prototype.fromString = function(s) {
    return this.fromXML($.parseXML(s));
}

// Elegir el color a mostrar según notas y actividades
Presentity.prototype.suggestStateColor = function() {
    var color = '#8cbe29';
    if (this.open) {
        if (this.activities.length == 1 && this.activities.indexOf('available') != -1) {
            // Estado disponible estilo Blink
            color = '#8cbe29';
        } else {
            if (this.activities.length > 0) color = 'orange';
            if (this.activities.indexOf('busy') != -1) color = 'red';
            if (this.activities.indexOf('on-the-phone') != -1) color = 'red';
        }
    } else {
        color = 'grey';
    }

    return color;
}


function SIPPresence(ua, gui)
{
    this._gui = gui;
    this.ua = ua;
    this.presentity = new Presentity();
    this.presentity.user = ua.configuration.authorizationUser;
    this.presentity.domain = ua.configuration.hostportParams;
    this.presenceETag = null;
    this.presenceTimer = null;
    this.publishRequest = null;
    this.roster = {}; // Subscripciones a lista de contactos
    this.subsWatch = null;  // Subscripción a presence.winfo

    /* Ejecutar la subscripción al evento presence.winfo . Este evento informa de
     * quién desea observar al usuario local, y la política seguida aquí es la
     * de autorizar de inmediato la escucha. */
    /*
    <?xml version="1.0"?>
    <watcherinfo xmlns="urn:ietf:params:xml:ns:watcherinfo" version="1" state="full">
      <watcher-list resource="sip:avillacis@pbx.villacis.com" package="presence"/>
    </watcherinfo>

    <?xml version="1.0"?>
    <watcherinfo xmlns="urn:ietf:params:xml:ns:watcherinfo" version="2" state="partial">
      <watcher-list resource="sip:avillacis@pbx.villacis.com" package="presence">
        <watcher id="87590e851228150e5980f7ca45a1b9ce@0:0:0:0:0:0:0:0" event="subscribe" status="pending">sip:gmacas@pbx.villacis.com</watcher>
      </watcher-list>
    </watcherinfo>

    */
    this._subscribeWithServerCheck(this.getLocalContact(), 'presence.winfo', function(contact, subscription) {
        this.subsWatch = subscription;
    }.bind(this), function(notification) {
        var xmlwatch = $.parseXML(notification.request.body);

        $(xmlwatch).find('watcherinfo > watcher-list[package=presence] watcher[event=subscribe][status=pending]')
            .each(function(idx, value) {
            //console.log("Aprobando ingreso a roster de contacto: " + $(value).text());
            this.subscribeToRoster($(value).text().replace(/^sip:/, ''));
        }.bind(this));
    }.bind(this));
}
SIPPresence.prototype.getLocalContact = function() {
    return this.ua.configuration.authorizationUser + '@' + this.ua.configuration.hostportParams;
}

/**
 * Método para iniciar la publicación de la presencia del usuario local.
 * Para mitigar la situación de que no es posible cerrar la presencia de
 * forma síncrona, se establece la expiración de la presencia a 90 segundos,
 * y se actualiza la presencia cada 60 segundos con un timer.
 */
SIPPresence.prototype.publishPresence = function () {

    if (this.presenceETag == null) {
        // Verificar si se tiene un PUBLISH previo en el servidor
        $.get('index.php', {
            menu: '_elastixpanel',
            elastixpanel: 'elastixwebphone',
            action: 'getPublishState'
        }, function(data) {
            if (data.message != "") {
                this.presenceETag = (data.message.ETag != "") ? data.message.ETag : null; // puede ser null
                this.presentity.note = data.message.note;
                this.presentity.activities = data.message.activities;
            }
            this._publishPresence();
        }.bind(this));
    } else {
        this._publishPresence();
    }

    var uris = this._gui.getContactURIs();
    var i;
    for (i = 0; i < uris.length; i++) this.subscribeToRoster(uris[i]);
}
SIPPresence.prototype._publishPresence = function() {
    if (this.publishRequest == null) {
        var extrahdr = [
            'Event: presence',
            'Content-Type: application/pidf+xml',
            'Contact: ' + this.ua.contact.toString()
        ];
        this.publishRequest = this.ua.request('PUBLISH', this.getLocalContact(), {
            body: this.presentity.toString(),
            extraHeaders: extrahdr
        });
        this.publishRequest.request.setHeader('Expires', '90');
        if (this.presenceETag != null) {
            this.publishRequest.request.setHeader('SIP-If-Match', this.presenceETag);
        }
        this.publishRequest.on('accepted', function (response, cause) {
            if (response.getHeader('Expires') != "0") {
                this.presenceETag = response.getHeader('Sip-Etag');
                // mandar this.presenceETag al servidor para recuperar
                // SIP-If-Match luego de recargar la página
                $.post('index.php', {
                    menu: '_elastixpanel',
                    elastixpanel: 'elastixwebphone',
                    action: 'setPublishState',
                    ETag: this.presenceETag,
                    note: this.presentity.note,
                    activities: this.presentity.activities
                }, function(data) {});

                this.publishRequest.request.setHeader('SIP-If-Match', this.presenceETag);

                // Elegir el color a mostrar según notas y actividades
                this._gui.updateMyStatus(this.presentity.suggestStateColor());
            } else {
                this.publishRequest = null;
                this.presenceETag = null;
                $.post('index.php', {
                    menu: '_elastixpanel',
                    elastixpanel: 'elastixwebphone',
                    action: 'setPublishState',
                    ETag: ''
                }, function(data) {});
                this._gui.updateMyStatus('gray');
            }
        }.bind(this));
        this.publishRequest.on('rejected', function (response, cause) {
            if (response.status_code == 412) {
                /* 412 Conditional request failed */
                if (this.presenceETag != null) {
                    /* Si ocurre este error, entonces ha ocurrido un error
                     * de sincronización, y el ETag almacenado es inválido.
                     */
                    this.presenceETag = null;
                    delete this.publishRequest.request.headers['Sip-If-Match'];
                    this.publishRequest.send();
                    return;
                }
            }

            // Ha ocurrido otro error que no se sabe manejar
            if (response.status_code == 489) {
                // Asterisk no soporta PUBLISH Event: presence
                console.debug(response.status_code + ' ' + cause);
            } else {
                this._gui.displayStatusMessage(
                    'Failed to PUBLISH presence: ' + response.status_code + ' ' + cause);
            }
        }.bind(this));
    } else {
        this.publishRequest.request.body = this.presentity.toString();
        this.publishRequest.send();
    }

    if (null == this.presenceTimer)
        this.presenceTimer = window.setInterval(this.publishPresence.bind(this), 60 * 1000);
}

/**
 * Método para retirar la presencia de la cuenta local, efectivamente indicado
 * que se retira la sesión de chat.
 */
SIPPresence.prototype.withdrawPresence = function () {

    if (null != this.presenceTimer) {
        window.clearInterval(this.presenceTimer);
        this.presenceTimer = null;
    }

    if (null == this.publishRequest) return;

    this.publishRequest.request.setHeader('Expires', '0');
    this.publishRequest.send();

    this._unsubscribeWithServerCheck(this.getLocalContact(), this.subsWatch);
    this.subsWatch = null;
    for (var contact in this.roster) this.unsubscribeFromRoster(contact);

}

SIPPresence.prototype.setPresenceStatus = function (note, activities) {
    this.presentity.note = note;
    this.presentity.activities = activities;
    this.publishPresence();
}

/**
 * Método privado que encapsula una subscripción a un contacto para que se
 * verifique si hay un Call-ID previo para una subscripción previa a ese
 * mismo contacto. Si la hay, la subscripción resultante renueva la
 * subscripción anterior en lugar de crear una nueva subscripción.
 */
SIPPresence.prototype._subscribeWithServerCheck = function(contact, event, checkCallback, notifyCallback) {
    var subscription;
    var s_options = {
        expires:        120,
        extraHeaders:   []
    };

    if (event == 'presence') {
        s_options.extraHeaders.push('Accept: application/pidf+xml');
    } else if (event == 'presence.winfo') {
        s_options.extraHeaders.push('Accept: application/watcherinfo+xml');
    }
    subscription = this.ua.subscribe(contact, event, s_options);
    subscription.on('notify', notifyCallback);
    checkCallback(contact, subscription);
}

/**
 * Método privado para deshacer una subscripción y anular el Call-ID de la
 * subscripción que se destruye.
 */
SIPPresence.prototype._unsubscribeWithServerCheck = function(contact, subscription) {
    if (subscription != null) {
        subscription.unsubscribe();
        subscription.close();
    }
}

/**
 * Método para subscribirse a la presencia de un contacto, y recibir estado
 * de presencia. Hasta ahora funciona con Jitsi.
 */
SIPPresence.prototype.subscribeToRoster = function(contact) {
    if (this.roster[contact] != null) return;

    this._subscribeWithServerCheck(contact, 'presence', function(contact, subscription) {
        this.roster[contact] = subscription;
    }.bind(this), function (notification) {
        var pres = new Presentity();
        pres.open = false;
        pres.note = 'Offline';
        pres.user = notification.request.from.uri.user;
        pres.domain = notification.request.from.uri.host;
        if ('' != notification.request.body) {
            pres.fromString(notification.request.body);
        }
        this._updateContactStatus(pres.user + "@" + pres.domain,
            pres.suggestStateColor(), pres.note);
    }.bind(this))
}

/**
 * Método para quitar la subscripción del contacto.
 */
SIPPresence.prototype.unsubscribeFromRoster = function(contact) {
    if (this.roster[contact] == null) return;
    this._unsubscribeWithServerCheck(contact, this.roster[contact]);
    delete this.roster[contact];

    this._updateContactStatus(contact, 'grey', '(unknown)');
}

SIPPresence.prototype._updateContactStatus = function(contact, newColor, newNote) {
    this._gui.updateContactStatus(contact, newColor, newNote);
}


function SIPBLF(ua, gui)
{
    this._gui = gui;
    this.ua = ua;
    this.roster = {};   // Subscripciones a BLF de contactos

    var uris = this._gui.getContactURIs();
    var i;
    for (i = 0; i < uris.length; i++) this.subscribeToRoster(uris[i]);
}
SIPBLF.prototype.subscribeToRoster = function(contact) {
    if (this.roster[contact] != null) return;

    var s_options = {
        expires:        120,
        extraHeaders:   [
            'Accept: application/dialog-info+xml'
        ]
    };
    this.roster[contact] = this.ua.subscribe(contact, 'dialog', s_options);
    this.roster[contact].on('notify', this._notifyBLF.bind(this));
}
SIPBLF.prototype.unsubscribeFromRoster = function(contact) {
    if (this.roster[contact] == null) return;

    this.roster[contact].unsubscribe();
    this.roster[contact].close();
    delete this.roster[contact];

    this._gui.updateContactBLF(contact, 'idle');
}
SIPBLF.prototype._notifyBLF = function(notification) {
    var xml_blf = $.parseXML(notification.request.body);
    var xml_dialog_info = xml_blf.getElementsByTagName('dialog-info')[0];
    var sipuri = decodeURIComponent(xml_dialog_info.getAttribute('entity'));
    var m = /^(sip:)?(\S+@\S+)$/.exec(sipuri);
    var contact = m[2];

    var xmlstate = xml_blf.getElementsByTagName('state');
    var nstatus = 'terminated';
    if (xmlstate.length > 0) {
        // Asume que el texto es el único contenido
        if (xmlstate[0].childNodes.length > 0)
            nstatus = xmlstate[0].childNodes[0].nodeValue;
    }
    switch (nstatus) {
    case 'trying':
    case 'proceeding':
    case 'early':
        this._gui.updateContactBLF(contact, 'ringing');
        break;
    case 'confirmed':
        this._gui.updateContactBLF(contact, 'oncall');
        break;
    default:
        this._gui.updateContactBLF(contact, 'idle');
        break;
    }
}

function PlayDTMF()
{
    this._freq = {
        '1': [697,1209],'2': [697,1336],'3': [697,1477],'A': [697,1633],
        '4': [770,1209],'5': [770,1336],'6': [770,1477],'B': [770,1633],
        '7': [852,1209],'8': [852,1336],'9': [852,1477],'C': [852,1633],
        '*': [941,1209],'0': [941,1336],'#': [941,1477],'D': [941,1633]
    };
    this._ctx = null;
    this.duration = 100;
    var cls = (window.AudioContext || window.webkitAudioContext);
    if (cls != undefined) {
        this._ctx = new cls();
    }
}
PlayDTMF.prototype.play = function(c) {
    if (this._ctx == null) return;
    if (typeof this._freq[c] == 'undefined') return;

    var o = this._freq[c].map(function(f) {
        var osc = this._ctx.createOscillator();
        osc.frequency.value = f;
        osc.connect(this._ctx.destination);
        osc.start();
        return osc;
    }, this);
    setTimeout(function() {
        o.forEach(function(osc) {
            osc.stop();
            osc.disconnect(this._ctx.destination)
        }, this);
    }.bind(this), this.duration);
}

